// +---------------------------------------------------------------------------+
//
//      Program:    hott_voltage_sensor.c
//
//      Purpose:    Reads the total voltage of a 3S LiPo battery in a RC plane
//					and reports that value back to the remote control, just
//					like a commercial Hott sensor.
//                  Only the sum over the 3 cells is measured, not every single cell
//      
//      Target:     ATtiny13A
//
//      Author:     Herbert Dingfelder, DL5NEG
//
//      License:    GPL 3
//
//      History:    13.12.2015  Initial Coding
//                  18.12.2015  Remote control show first reaction to sensor
//                              (still with static values taken from commercial sensor)
//					19.12.2015  All parameter tried out and works great -> the problem
//                              to get a homemade HoTT protocol sensor to work is solved
//					20.12.2015	Tested on the lab-bench, works great in all implemented feature
//								=> Version 1.0 released
//
// +---------------------------------------------------------------------------+

#define F_CPU 9600000
#include <util/delay.h>
#include <avr/io.h>
#include <stdio.h>
#include <string.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/sleep.h>
#include <avr/cpufunc.h>

#define DEVICE_SMD_1				// specify here for which hardware sample the compile is

// *** finetune the OSCCAL value for every ATtiny13A device! ***
// *** in function eceive_byte_via_sw_uart() uncomment the pulse output
// *** on the LED pin, then aim for 416µs for 8 fullwave pulses 
#ifdef DEVICE_PROTOTYPE
	#define OSCCAL_DEVICE_SPECIFIC	58	// this value is for the DIL8 prototype
#endif
#ifdef DEVICE_SMD_1
	#define OSCCAL_DEVICE_SPECIFIC	56	// this value is for the first SMD device
#endif





#define LOW_VOLTAGE_THRESHOLD	102	// *** chose your low voltage waring level, e.g. 102 for 10.2V = 3.4V per cell (100mV steps)

//#define DEBUG_OUT					// *** set this define to activate debug output (needs about 100 byte flash) *** 



#define HOTT_PORT PORTB				// HoTT interface is connected to PB4
#define HOTT_DDR  DDRB
#define HOTT_IN   PINB
#define HOTT_BIT  _BV(4)

#define TXD_PORT HOTT_PORT          // TXD and RXD are both connected to the one HoTT pin
#define TXD_PIN  HOTT_BIT
#define TXD_DDR  HOTT_DDR
#define RXD_PIN  HOTT_BIT
#define RXD_READ HOTT_IN

#define LIPO_ADC_CHANNEL 3          // LiPo battery is connected via divider to ADC 3          

#define LED_PORT PORTB				// LED is connected to PB1
#define LED_DDR  DDRB
#define LED_BIT  _BV(1)

#define DEBUG_PORT PORTB			// debug pin is PB 0
#define DEBUG_DDR  DDRB
#define DEBUG_PIN  _BV(0)


// ----- macros to switch TXD to low/high, to enhance code readability -----

#define UART_TXD_LOW    (TXD_PORT &= ~(TXD_PIN) )       // macro to set TXD low
#define UART_TXD_HIGH   (TXD_PORT |=  (TXD_PIN) )       // macro to set TXD high
#define UART_OPMODE_TX	(TXD_DDR |=   (TXD_PIN) ) 		// macro to switch UART pin to output
#define UART_OPMODE_RX	(TXD_DDR &=  ~(TXD_PIN) ) 		// macro to switch UART pin to input
#define UART_RXD        (RXD_READ & (RXD_PIN) )         // macro to read RXD back
#define LED_ON          (LED_PORT |=  LED_BIT)          // macro to switch on LED
#define LED_OFF         (LED_PORT &= ~LED_BIT)          // macro to switch off LED
#define DEBUG_OUT_LOW	(DEBUG_PORT &= ~(DEBUG_PIN))	// macro to set debug output low
#define DEBUG_OUT_HIGH	(DEBUG_PORT |=  (DEBUG_PIN))	// macro to set debug output high

void reset_handler(void);
unsigned int read_adc(unsigned char adc_number);
unsigned char hex2asc(unsigned char hexbyte);
void send_byte_via_sw_uart(char byte_to_send);
int receive_byte_via_sw_uart(void);
void debug_output(uint16_t value);
uint16_t read_vbatt(void);
void wait_for_next_uart_cycle(void);
void init_RS232_timer(void);

// global variable checksum (helps so safe memory to have this global)
uint16_t checksum;

// ============================= main program starts here ======================
int main(void)
{
      
	uint16_t adc_result, debug_dummy=12345;
    int fuel_calc;
    uint8_t received_databyte;
    uint8_t battery_voltage=124;                        // default for battery above warning threshold
    uint8_t cell_voltage=200;                           // default for cell voltage within a normal range
    uint8_t temperature=20+25;                          // default for temperature is 25°C (20 offset per protocol definiton)
    uint8_t fuel_bar=50;                                // default for fuel bar graph is 50%
    uint8_t alarm_beep = 0x00;                          // default for alarm beep is off
    
    
    reset_handler();                                    // do all the regular port and register settings
                                                        //   and start the needed UART counter
    
	for(;;)                                             // main loop starts here
	{

		received_databyte = receive_byte_via_sw_uart();	// wait for next character on Hott interface
		
		if(received_databyte != 0x80 )					// if it is not the first byte of our address
			continue;									//   start a new loop immediately

		received_databyte = receive_byte_via_sw_uart();	// wait for next character on Hott interface
		
		if(received_databyte != 0x8d )					// if it is not the second byte of our address
			continue;									//   start a new loop immediately
	
		// now we know we have been called and shall respond with our information
		
		LED_ON;                                         // have the LED on while we transmit on HoTT line
		
		_delay_ms(3);                                   // protocol needs 5ms, we have 2ms more in send_byte_via_sw_uard()
		
		UART_OPMODE_TX;                                 // switch HoTT pin to output mode

        checksum = 0;                                   // reset checksum at start of HoTT telegram

		send_byte_via_sw_uart(0x7c);					// 1 start byte constant 0x7c
		send_byte_via_sw_uart(0x8d);					// 2 sensor id, 0x8d = general module
		send_byte_via_sw_uart(alarm_beep);				// 3 warning_beeps (triggers respective voice output), 0x10= low voltage, 0x00 = no alarm => tested ok
		send_byte_via_sw_uart(0xd0);					// 4 constant value 0xd0
		send_byte_via_sw_uart(0x00);					// 5 alarm bitmask inverted, 0 all cell voltage, 1 Battery 1, 2 Battery 2, 3 Temperature 1, 4 Temperature 2, 5 Fuel, 6 mAh, 7 Altitude
		send_byte_via_sw_uart(0x00);					// 6 alarm bitmask inverted, 0 main power current, 1 main power voltage, 2 Altitude, 3 m/s, 4 m/3s// 5 unknown// 6 unknown// 7 "ON" sign/text msg active
		send_byte_via_sw_uart(cell_voltage);			// 7 voltage cell 1 in 20mV => tested ok
		send_byte_via_sw_uart(cell_voltage);			// 8 voltage cell 2 in 20mV => tested ok
		send_byte_via_sw_uart(cell_voltage);			// 9 voltage cell 3 in 20mV => tested ok
		send_byte_via_sw_uart(0x00);					// 10 voltage cell 4 in 20mV => tested ok
		send_byte_via_sw_uart(0x00);					// 11 voltage cell 5 in 20mV => tested ok
		send_byte_via_sw_uart(0x00);					// 12 voltage cell 6 in 20mV => tested ok
		send_byte_via_sw_uart(battery_voltage);			// 13 overall voltage battery 1 in 100mV (LSB) => tested ok
		send_byte_via_sw_uart(0x00);					// 14 overall voltage battery 1 in 100mV (MSB) => tested ok (set to 0x00 here, as a 3S battery does not reach 25.6V)
		send_byte_via_sw_uart(0x00);					// 15 overall voltage battery 2 in 100mV (LSB)  => tested ok
		send_byte_via_sw_uart(0x01);					// 16 overall voltage battery 2 in 100mV (MSB)   => tested ok
		send_byte_via_sw_uart(temperature);				// 17 temperature 1 (offset 20, 20=0°C) => tested ok
		send_byte_via_sw_uart(0x00);					// 18 temperature 2 (offset 20, 20=0°C) => tested ok
		send_byte_via_sw_uart(fuel_bar);				// 19 fuel shown as % in bargraph (0..100) => tested ok
		#ifdef DEBUG_OUT
			send_byte_via_sw_uart( debug_dummy    &0x00ff);	// the fuel value is very convenient for debugging, at it is shown as number 
			send_byte_via_sw_uart((debug_dummy>>8)&0x00ff);	//   in the display of the LCD, we assign whatever 16bit value we are interessted in to monitor
		#else
			send_byte_via_sw_uart(fuel_bar);			// 20 fuel shown in ml as text (full=65535) LSB, we use it as % number => tested ok (sidenote: value stalls at 65000) 
			send_byte_via_sw_uart(0x00);				// 21 fuel shown in ml as text (full=65535) MSB => tested ok (see line above)
		#endif // DEBUG_OUT
		send_byte_via_sw_uart(0x00);    				// 22 rpm in steps of 10 (100=1000rpm) LSB => tested ok
		send_byte_via_sw_uart(0x00);					// 23 rpm in steps of 10 (100=1000rpm) MSB => tested ok
		send_byte_via_sw_uart(0xf4);					// 24 altitude in m (500 offset, 0x01f4=500=0m) LSB => tested ok (result only shown as text)
		send_byte_via_sw_uart(0x01);					// 25 altitude in m (500 offset, 0x01fa=500=0m) MSB => tested ok (result only shown as text)
		send_byte_via_sw_uart(0x30);					// 26 climb rate in 0.01ms/s, 0x7530=30000=0.00m/s LSB => tested ok
		send_byte_via_sw_uart(0x75);					// 27 climb rate in 0.01ms/s, 0x7530=30000=0.00m/s MSB => tested ok
		send_byte_via_sw_uart(0x78);                    // 28 climb rate in m/3s, 0x78=120=0m/3s => tested ok
		send_byte_via_sw_uart(0x00);					// 29 current in 100mA, 100=10A LSB => tested ok
		send_byte_via_sw_uart(0x00);					// 30 current in 100mA, 100=10A MSB => tested ok
		send_byte_via_sw_uart(battery_voltage);			// 31 main power voltage in 100mV (shown in overview screen) LSB => tested ok
		send_byte_via_sw_uart(0x00);					// 32 main power voltage in 100mV (shown in overview screen) MSB => tested ok (set to 0x00 here, as a 3S battery does not reach 25.6V)
		send_byte_via_sw_uart(0x00);					// 33 used battery capacity in 10mAh, LSB => tested ok
		send_byte_via_sw_uart(0x00);					// 34 used battery capacity in 10mAh, MSB => tested ok
		send_byte_via_sw_uart(0x00);					// 35 speed LSB => not tested
		send_byte_via_sw_uart(0x00);					// 36 speed MSB => not tested
		send_byte_via_sw_uart(0x00);					// 37 minimum cell voltage in 20mV steps? => not tested
		send_byte_via_sw_uart(0x00);					// 38 number of the cell with the lowest voltage => not tested
		send_byte_via_sw_uart(0x00);					// 39 rpm 2 in steps of 10 (100=1000rpm) LSB => not tested
		send_byte_via_sw_uart(0x00);					// 40 rpm 2 in steps of 10 (100=1000rpm) MSB => not tested
		send_byte_via_sw_uart(0x00);					// 41 general error number (12=voice) => not tested
		send_byte_via_sw_uart(0x00);					// 42 high pressure up to 16bar in 100mbar  => not tested
		send_byte_via_sw_uart(0x00);					// 43 version number => not tested
		send_byte_via_sw_uart(0x7d);					// 44 stop byte 0x7d
        send_byte_via_sw_uart(checksum & 0xFF);			// 45 checksum (last 8bits of sum of all data bytes)
		
		_delay_ms(1);                                   // pause of 1ms before we switch back to receive mode
		
        UART_OPMODE_RX;                                 // switch back to receive mode
		
		LED_OFF;                                        // switch LED off, we are done with transmitting
		
        // now we can take care to get new measurement values from the ADC

        adc_result = read_vbatt();                      // read the LiPo battery overall voltage via the ADC

        battery_voltage = (adc_result * 36) / 256;      // Vbatt in 100mV, tricky way to calculate, to safe memory
    
        // for this simple module, we fake the cell voltage by dividing the overall voltage by 3 (for 3s LiPo)
        cell_voltage = (adc_result * 60) / 256;         // Vcell in 20mV, tricky way to calculate, to safe memory
 
        // we use the fuel bar graph to show the percentage of remaining battery capacity in range 3.2 to 4.2V
        // (again we use very memory-optimized mathematics, with only division that the compiler can
        //   do as shift right operations and only multiplication in the 16bit range)
        fuel_calc = (int)adc_result-684;                // remove offset (ADC value at 3.2V per cell)
        if(fuel_calc<0)									// suppress underflow below 0%
			fuel_calc=0;
        fuel_calc *= 15;                                // *15
        fuel_calc >>= 5;                                // /16  => *15/32 is the factor we need to get 100% at 4.2V
        fuel_bar = fuel_calc & 0x00ff;
		if(battery_voltage < LOW_VOLTAGE_THRESHOLD)		// if voltage too low
			alarm_beep = 0x10;							// trigger acoustic warning at remote control
		else
			alarm_beep = 0x00;							// no warning when voltage is ok
    
    }

    return 0;
    
}



// --------------------------------- reset handler -----------------------------
void reset_handler(void)
{
    
    CLKPR = 0x80;                       // set clock prescaler change enable 
    CLKPR = 0x00;                       // set prescaler to clock division factor 1
                                        // => system clock runs at the full 9.6MHz
    
	OSCCAL = OSCCAL_DEVICE_SPECIFIC;    // this value needs to be trimmed for every sample of ATtiny13A*
    
	init_RS232_timer();				    // start a clock counter for 19k2 baudrate
     
	LED_DDR |= LED_BIT;				    // set LED pin as output
    LED_OFF;                            // start with LED off
	
	UART_OPMODE_RX;					    // the HoTT interface pin to high impedance first (receiving)
    UART_TXD_HIGH;					    // keep the logical value on the HoTT high from the start, otherwise
                                        //   switching to TX mode will cause a falling edge right away which
                                        //   would erronously be detected as 0x00 data byte the the HoTT receiver

	#ifdef DEBUG_OUT				
		DEBUG_DDR |= DEBUG_PIN; 	    // set debug pin as output 
		DEBUG_OUT_HIGH;				    // and set it high (default for serial communication)
	#endif	// DEBUG_OUT

}

//------------------------------- read adc -----------------------
unsigned int read_adc(unsigned char adc_number)
{

    uint16_t adc_result;
    
    ADCSRA |= _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0); // ADC prescaler :128, that 
                                        // gives ADC frequency of 11.059MHz/128 = 86kHz

    ADMUX = adc_number;                 // select Vcc reference (Bits 7 and 6 remain zero)
                                        // and chose the ADC channel

    ADCSRA |= _BV(ADEN);                // switch on the ADC in general

    ADCSRA |= _BV(ADSC);                // start a single ADC conversion
    while( ADCSRA & _BV(ADSC) );        // wait until conversion is complete

    // data sheet recommends to discard first conversion, so we do one more
    ADCSRA |= _BV(ADSC);                // start a single ADC conversion
    while( ADCSRA & _BV(ADSC) );        // wait until conversion is complete

    adc_result = ADC;                   // take over ADC reading result into variable

    ADCSRA = 0;                         // completely disable the ADC to save power
    	
	return adc_result;
	
}

#ifdef DEBUG_OUT
//--------------------------- hex to asc ---------------------------------
// converts a hex value (00..0F) into the ASCII-character '0'..'F'
unsigned char hex2asc(unsigned char hexbyte)
{
    
    hexbyte += 48;                      // ASCII code for 0-9 is 48-57
    
    if(hexbyte>=58)                     // ASCII code for A-F is 65-70 -> in that case add 7
        hexbyte += 7;
    
    return hexbyte;
}
#endif


// ------------------------------- send byte via uart --------------------------
void send_byte_via_sw_uart(char byte_to_send)
{
    uint8_t bit;    
    
    _delay_ms(2);						// the Hott protocol wants delays between data bytes,
										//   we place this delay generically here rather than
										//	 for every call of this function, to save memory
    
    checksum += byte_to_send;           // to add up the checksum for the complete protocol here
                                        //   is a way to safe memory
    
    wait_for_next_uart_cycle();         // synchronize with the next overflow of the timer
    
    UART_TXD_LOW;                       // send startbit (always 0)
    
    #ifdef DEBUG_OUT
		DEBUG_OUT_LOW;
    #endif	// DEBUG_OUT
    
    for( bit=0 ; bit<8 ; bit++ )        // send bit by bit, starting with LSB
    {
       
        wait_for_next_uart_cycle();     // synchronize with the next overflow of the timer
        
        if(byte_to_send & _BV(0))       // take TXD line to high if LSB is set
        {
            UART_TXD_HIGH;
            #ifdef DEBUG_OUT
				DEBUG_OUT_HIGH;
			#endif	// DEBUG_OUT
        }
        else
        {
            UART_TXD_LOW; 				// respectively take it low if LSB is not set
            #ifdef DEBUG_OUT
				DEBUG_OUT_LOW;
			#endif	// DEBUG_OUT           
        }
        
        byte_to_send>>=1;               // shift the byte to be send by one bit towards LSB
    
    }   
    
    wait_for_next_uart_cycle();         // synchronize with the next overflow of the timer

    UART_TXD_HIGH;                      // send stopbit (always 1)
    
    #ifdef DEBUG_OUT
		DEBUG_OUT_HIGH;
     #endif	// DEBUG_OUT
       
}


// ------------------------------ receive byte via uart (blocking)------------------------
int receive_byte_via_sw_uart(void)
{
  
    unsigned char bit_counter;
    unsigned char byte_received=0;
  
	UART_OPMODE_RX;								// switch the joint UART pin to RX mode
    
    while( UART_RXD );                          // wait for UART RXD to go low (beginning of startbit)
												// (this is the most frequented waiting loop)
    
    TCNT0 = 0;                                  // reset counter to synchronize on incoming bits
	
                                                // read 8 data bit (plus start bit)
    for( bit_counter = 0; bit_counter < 9 ; bit_counter++ )
    {

        byte_received >>= 1;                    // shift the so far received bits one right
            
        TIFR0 |= _BV(OCF0B);                    // clear output compare B match flag
        while( !(TIFR0 & _BV(OCF0B)) );         // wait for middle of bit
          
        if( UART_RXD )                          // sample on RXD, if it is high...
        {
            byte_received |= _BV(7);            // ...set MSB in received byte
        }
        
        LED_ON; _delay_us(20); LED_OFF;		// helps to finetune OSCCAL (aim for 416µs for 8 pulses)
        
    }

    // now only the stop bit is left, which we simply ignore as it is always high
    TIFR0 |= _BV(OCF0B);                    	// clear output compare B match flag
    while( !(TIFR0 & _BV(OCF0B)) );         	// wait for middle of bit

    // intentionally we do not wait for the end of the stop bit here, that gives us
    // half a bit of time to process the received byte, before the next one may arive
    // (waiting for end of stop bit could be done with: wait_for_next_uart_cycle(); )

    return byte_received;
    
}

#ifdef DEBUG_OUT
// --- send the 16bit parameter value as readable hex number to UART for debugging ---
void debug_output(uint16_t value)
{
    
        send_byte_via_sw_uart('0');
        send_byte_via_sw_uart('x');
        send_byte_via_sw_uart(hex2asc((value>>12)&0x0f));
        send_byte_via_sw_uart(hex2asc((value>> 8)&0x0f));
        send_byte_via_sw_uart(hex2asc((value>> 4)&0x0f));
        send_byte_via_sw_uart(hex2asc((value    )&0x0f));
        send_byte_via_sw_uart(' ');    

}
#endif


// ----- reads the battery voltage via a 1:4.3 voltage divider at 3.34V reference, with averaging, returns voltage in mV -----
uint16_t read_vbatt(void)
{

    uint16_t adc_result, avg_sum; 
    uint8_t a;
   
    avg_sum=0;
    
    for(a=0; a<8; a++)                              // do an averaging over 8 ADC reads
    {
        adc_result = read_adc(LIPO_ADC_CHANNEL);    // read out the ADC that is connected to Vbat
        
        avg_sum += adc_result;                      // add up the results
       
        _delay_ms(1);                               // wait 1ms for good independenc between single reads
    }
    
    avg_sum >>= 3;                                  // divide the added results by 8 for final result
    
    return avg_sum;                                 // return the battery voltage as raw value (0..0x3FF)

}


// ----------- synchronize with the next overflow of timer1 -----------------
void wait_for_next_uart_cycle(void)
{
    
    TIFR0 |= _BV(OCF0A);                            // clear the timer overflow flag
    while( !( TIFR0 & _BV(OCF0A)) );                // wait until overflow flag is set again

}

// --- switch on a permanent clock source that runs at the baudrate of 19200kHz
void init_RS232_timer(void)
{
    
    // baudrate is 19200 bit/s => bit length is 52.08µs
    // main clock is 9.6MHz, with prescaler 8 => clock period 833.33ns 
    // => 62.4998 clock cycles per bit length
    
    // set counter compare match to 62.5 -1 is roughly 62
    // (the -1 is due to the fact that the value is actually reached before the counter is reset)
    OCR0A = 62;        // compiler can handle the 16bit register access, simply assign the number
    
    // for sampling incoming bits in the middle, use the second compare match unit at half that value
    OCR0B = 31;
    
    // switch on the counter with prescale factor 8 
    // and switch on the clear-on-compare-match feature to reset the counter automatically when the
    // value of the compare match register A is reached
    TCCR0A = _BV(WGM01);  				// clear timer on compare with OCR0A
    TCCR0B = _BV(CS01);					// clock prescaler is 8
 
    TCNT0 = 0;                          // reset the counter, so it starts to count at zero
    
}


